/*
    OCLTools is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Dr. Zaius 
    ClusterChimps.org
*/

// standard utilities and systems includes
#include <stdio.h>
#include <vector>
#include <string>
#include <iostream>
#include <CL/cl.h>
#include <oclutil.h>
#include <CL/cl_ext.h>

#include <boost/program_options/options_description.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/parsers.hpp>

static int nGpuArchCoresPerSM[] = { -1, 8, 32 };

extern "C" void ocltLogBuildInfo(cl_program cpProgram, cl_device_id cdDevice);

const char* version = "oclcc 1.0\nCopyright (C) 2011 ClusterChimps.org\nThis software \
is free (GPLv3). There is NO warranty; not even \nfor MERCHANTABILITY or FITNESS \
FOR A PARTICULAR PURPOSE.\n";

const char* info = "This program is a stand alone (offline) OpenCL compiler.  It accepts all\n\
OpenCL compiler flags and Nvidia extensions.  This compiler is intended to\n\
be used as part of the OCLTools suite of development tools and may not function\n\
correctly outside of the OCLTools suite of tools.\n";


template < class T >
std::string ToString(const T &arg)
{
        std::ostringstream   out;
        out << arg;

        return(out.str());
}

void getDeviceInfo(std::string &output, cl_device_id device)
{
   char info[1024];
   bool nv_device_attibute_query = false;

   clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(info), &info, NULL);
   output += "  CL_DEVICE_NAME:\t\t\t" + std::string(info) + "\n";

   clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(info), &info, NULL);
   output += "  CL_DEVICE_VENDOR:\t\t\t" + std::string(info) + "\n";

   clGetDeviceInfo(device, CL_DRIVER_VERSION, sizeof(info), &info, NULL);
   output += "  CL_DRIVER_VERSION:\t\t\t" + std::string(info) + "\n";

   cl_device_type type;
   clGetDeviceInfo(device, CL_DEVICE_TYPE, sizeof(type), &type, NULL);
   if( type & CL_DEVICE_TYPE_CPU )
       output += "  CL_DEVICE_TYPE:\t\t\tCL_DEVICE_TYPE_CPU\n";
   if( type & CL_DEVICE_TYPE_GPU )
       output += "  CL_DEVICE_TYPE:\t\t\tCL_DEVICE_TYPE_GPU\n";
   if( type & CL_DEVICE_TYPE_ACCELERATOR )
       output += "  CL_DEVICE_TYPE:\t\t\tCL_DEVICE_TYPE_ACCELERATOR\n";
   if( type & CL_DEVICE_TYPE_DEFAULT )
       output += "  CL_DEVICE_TYPE:\t\t\tCL_DEVICE_TYPE_DEFAULT\n";

   cl_uint compute_units;
   clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(compute_units), &compute_units, NULL);
   output += "  CL_DEVICE_MAX_COMPUTE_UNITS:\t\t" + ToString(compute_units) + "\n";

   size_t workitem_dims;
   clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof(workitem_dims), &workitem_dims, NULL);
   output += "  CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS:\t" + ToString((int)workitem_dims) + "\n";

   size_t workitem_size[3];
   clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(workitem_size), &workitem_size, NULL);
   output += "  CL_DEVICE_MAX_WORK_ITEM_SIZES:\t" + ToString((int)workitem_size[0]) + " / " +
             ToString((int)workitem_size[1]) + " / " + ToString((int)workitem_size[2]) + "\n";

   size_t workgroup_size;
   clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(workgroup_size), &workgroup_size, NULL);
   output += "  CL_DEVICE_MAX_WORK_GROUP_SIZE:\t" + ToString((int)workgroup_size) + "\n";

   cl_uint clock_frequency;
   clGetDeviceInfo(device, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(clock_frequency), &clock_frequency, NULL);
   output += "  CL_DEVICE_MAX_CLOCK_FREQUENCY:\t" + ToString((int)clock_frequency) + " MHz\n";

    cl_uint addr_bits;
   clGetDeviceInfo(device, CL_DEVICE_ADDRESS_BITS, sizeof(addr_bits), &addr_bits, NULL);
   output += "  CL_DEVICE_ADDRESS_BITS:\t\t" + ToString((int)addr_bits) + "\n";

   cl_ulong max_mem_alloc_size;
   clGetDeviceInfo(device, CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof(max_mem_alloc_size), &max_mem_alloc_size, NULL);
   output += "  CL_DEVICE_MAX_MEM_ALLOC_SIZE:\t\t" + ToString((ulong)max_mem_alloc_size/1024/1024) + " MB\n";

   cl_ulong mem_size;
   clGetDeviceInfo(device, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(mem_size), &mem_size, NULL);
   output += "  CL_DEVICE_GLOBAL_MEM_SIZE:\t\t" + ToString((ulong)mem_size/1024/1024) + " MB\n";

   cl_bool error_correction_support;
   std::string yes = "Yes";
   clGetDeviceInfo(device, CL_DEVICE_ERROR_CORRECTION_SUPPORT, sizeof(error_correction_support), &error_correction_support, NULL);
   if(error_correction_support == 0) yes = "No";
   output += "  CL_DEVICE_ERROR_CORRECTION_SUPPORT:\t" + yes + "\n";

   cl_device_local_mem_type local_mem_type;
   std::string memType = "local";
   clGetDeviceInfo(device, CL_DEVICE_LOCAL_MEM_TYPE, sizeof(local_mem_type), &local_mem_type, NULL);
   if(local_mem_type != 1) memType = "global";
   output += "  CL_DEVICE_LOCAL_MEM_TYPE:\t\t" + memType + "\n";

   clGetDeviceInfo(device, CL_DEVICE_LOCAL_MEM_SIZE, sizeof(mem_size), &mem_size, NULL);
   output += "  CL_DEVICE_LOCAL_MEM_SIZE:\t\t" + ToString((int)mem_size/1024) + "KB\n";

   clGetDeviceInfo(device, CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE, sizeof(mem_size), &mem_size, NULL);
   output += "  CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE:\t" + ToString((int)mem_size/1024) + "KB\n";

   cl_command_queue_properties queue_properties;
   clGetDeviceInfo(device, CL_DEVICE_QUEUE_PROPERTIES, sizeof(queue_properties), &queue_properties, NULL);
   if( queue_properties & CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE )
      output += "  CL_DEVICE_QUEUE_PROPERTIES:\t\tCL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE\n";
   if( queue_properties & CL_QUEUE_PROFILING_ENABLE )
      output += "  CL_DEVICE_QUEUE_PROPERTIES:\t\tCL_QUEUE_PROFILING_ENABLE\n";

    cl_bool image_support;
   clGetDeviceInfo(device, CL_DEVICE_IMAGE_SUPPORT, sizeof(image_support), &image_support, NULL);
   output += "  CL_DEVICE_IMAGE_SUPPORT:\t\t" + ToString((bool)image_support) + "\n";

   cl_uint max_read_image_args;
   clGetDeviceInfo(device, CL_DEVICE_MAX_READ_IMAGE_ARGS, sizeof(max_read_image_args), &max_read_image_args, NULL);
   output += "  CL_DEVICE_MAX_READ_IMAGE_ARGS:\t" + ToString((uint)max_read_image_args) + "\n";

   cl_uint max_write_image_args;
   clGetDeviceInfo(device, CL_DEVICE_MAX_WRITE_IMAGE_ARGS, sizeof(max_write_image_args), &max_write_image_args, NULL);
   output += "  CL_DEVICE_MAX_WRITE_IMAGE_ARGS:\t" + ToString((uint)max_write_image_args) + "\n";

   cl_device_fp_config fp_config;
   clGetDeviceInfo(device, CL_DEVICE_SINGLE_FP_CONFIG, sizeof(cl_device_fp_config), &fp_config, NULL);
   output += "  CL_DEVICE_SINGLE_FP_CONFIG:\t\t";
   if(fp_config & CL_FP_DENORM)           output += "denorms ";
   if(fp_config & CL_FP_INF_NAN)          output += "INF-quietNaNs ";
   if(fp_config & CL_FP_ROUND_TO_NEAREST) output += "round-to-nearest ";
   if(fp_config & CL_FP_ROUND_TO_ZERO)    output += "round-to-zero ";
   if(fp_config & CL_FP_ROUND_TO_INF)     output += "round-to-inf ";
   if(fp_config & CL_FP_FMA)              output += "fma ";
   output += "\n";

   size_t szMaxDims[5];
   clGetDeviceInfo(device, CL_DEVICE_IMAGE2D_MAX_WIDTH, sizeof(size_t), &szMaxDims[0], NULL);
   output +="  CL_DEVICE_IMAGE <dim>\t\t\t2D_MAX_WIDTH\t" + ToString(szMaxDims[0]) + "\n";

   clGetDeviceInfo(device, CL_DEVICE_IMAGE2D_MAX_HEIGHT, sizeof(size_t), &szMaxDims[1], NULL);
   output +="  \t\t\t\t\t2D_MAX_HEIGHT\t" + ToString(szMaxDims[1]) + "\n";

   clGetDeviceInfo(device, CL_DEVICE_IMAGE3D_MAX_WIDTH, sizeof(size_t), &szMaxDims[2], NULL);
   output +="  \t\t\t\t\t3D_MAX_WIDTH\t" + ToString(szMaxDims[2]) + "\n";

   clGetDeviceInfo(device, CL_DEVICE_IMAGE3D_MAX_HEIGHT, sizeof(size_t), &szMaxDims[3], NULL);
   output +="  \t\t\t\t\t3D_MAX_HEIGHT\t" + ToString(szMaxDims[3]) + "\n";

   clGetDeviceInfo(device, CL_DEVICE_IMAGE3D_MAX_DEPTH, sizeof(size_t), &szMaxDims[4], NULL);
   output +="  \t\t\t\t\t3D_MAX_DEPTH\t" + ToString(szMaxDims[4]) + "\n";

    clGetDeviceInfo(device, CL_DEVICE_EXTENSIONS, sizeof(info), &info, NULL);
   if (info != 0)
   {
      output +="  CL_DEVICE_EXTENSIONS:";
      std::string devString;
      devString = std::string(info);

      size_t szOldPos = 0;
      size_t szSpacePos = devString.find(' ', szOldPos); // extensions string is space delimited
      while (szSpacePos != devString.npos)
      {
          if( strcmp("cl_nv_device_attribute_query", devString.substr(szOldPos, szSpacePos - szOldPos).c_str()) == 0 )
              nv_device_attibute_query = true;

          if (szOldPos > 0)
          {
              output += "\t\t";
          }
          output += "\t\t\t" + devString.substr(szOldPos, szSpacePos - szOldPos) + "\n";

          do {
              szOldPos = szSpacePos + 1;
              szSpacePos = devString.find(' ', szOldPos);
          } while (szSpacePos == szOldPos);
      }

   } else {
      output +="  CL_DEVICE_EXTENSIONS\t\tnone";

   }

   if(nv_device_attibute_query)
   {
       cl_uint compute_capability_major, compute_capability_minor;
       clGetDeviceInfo(device, CL_DEVICE_COMPUTE_CAPABILITY_MAJOR_NV, sizeof(cl_uint), &compute_capability_major, NULL);
       clGetDeviceInfo(device, CL_DEVICE_COMPUTE_CAPABILITY_MINOR_NV, sizeof(cl_uint), &compute_capability_minor, NULL);
        output +="  CL_DEVICE_COMPUTE_CAPABILITY_NV\t" + ToString(compute_capability_major) +
                "." + ToString(compute_capability_minor) + "\n";

       output +="  NUMBER OF MULTIPROCESSORS\t\t" + ToString(compute_units) + "\n";
       output +="  NUMBER OF CUDA CORES\t\t\t" + ToString(nGpuArchCoresPerSM[compute_capability_major] * compute_units) + "\n";

       cl_uint regs_per_block;
       clGetDeviceInfo(device, CL_DEVICE_REGISTERS_PER_BLOCK_NV, sizeof(cl_uint), &regs_per_block, NULL);
       output +="  CL_DEVICE_REGISTERS_PER_BLOCK_NV\t" + ToString(regs_per_block) + "\n";

       cl_uint warp_size;
       clGetDeviceInfo(device, CL_DEVICE_WARP_SIZE_NV, sizeof(cl_uint), &warp_size, NULL);
       output +="  CL_DEVICE_WARP_SIZE_NV\t\t" + ToString(warp_size) + "\n";

       cl_bool gpu_overlap;
       clGetDeviceInfo(device, CL_DEVICE_GPU_OVERLAP_NV, sizeof(cl_bool), &gpu_overlap, NULL);
       yes = "CL_TRUE";
       if(gpu_overlap == 0) yes = "CL_FALSE";
       output +="  CL_DEVICE_GPU_OVERLAP_NV\t\t" + yes + "\n";

       cl_bool exec_timeout;
       clGetDeviceInfo(device, CL_DEVICE_KERNEL_EXEC_TIMEOUT_NV, sizeof(cl_bool), &exec_timeout, NULL);
       yes = "CL_TRUE";
       if(exec_timeout == 0) yes = "CL_FALSE";
       output +="  CL_DEVICE_KERNEL_EXEC_TIMEOUT_NV\t" + yes + "\n";

       cl_bool integrated_memory;
       clGetDeviceInfo(device, CL_DEVICE_INTEGRATED_MEMORY_NV, sizeof(cl_bool), &integrated_memory, NULL);
       yes = "CL_TRUE";
       if(integrated_memory == 0) yes = "CL_FALSE";
       output +="  CL_DEVICE_INTEGRATED_MEMORY_NV\t" + yes + "\n";

       cl_uint vec_width [6];
       clGetDeviceInfo(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_CHAR, sizeof(cl_uint), &vec_width[0], NULL);
       clGetDeviceInfo(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_SHORT, sizeof(cl_uint), &vec_width[1], NULL);
       clGetDeviceInfo(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_INT, sizeof(cl_uint), &vec_width[2], NULL);
       clGetDeviceInfo(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_LONG, sizeof(cl_uint), &vec_width[3], NULL);
        clGetDeviceInfo(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_FLOAT, sizeof(cl_uint), &vec_width[4], NULL);
       clGetDeviceInfo(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_DOUBLE, sizeof(cl_uint), &vec_width[5], NULL);
       output +="  CL_DEVICE_PREFERRED_VECTOR_WIDTH_<t>\t";
       output +="CHAR "    + ToString(vec_width[0]);
       output +=" SHORT "  + ToString(vec_width[1]);
       output +=" INT "    + ToString(vec_width[2]);
       output +=" LONG "   + ToString(vec_width[3]);
       output +=" FLOAT "  + ToString(vec_width[4]);
       output +=" DOUBLE " + ToString(vec_width[5]);
       output +="\n\n";

   }



}

void deviceQuery()
{
    std::string output = "Platform Info:\n";

    cl_platform_id clSelectedPlatformID = NULL;
    cl_int status = ocltGetPlatformID (&clSelectedPlatformID, "NVIDIA");
    checkError(status, CL_SUCCESS);

    char buffer[2048];
    // Get OpenCL platform name and version
    status = clGetPlatformInfo(clSelectedPlatformID, CL_PLATFORM_NAME, sizeof(buffer), buffer, NULL);
    if (status == CL_SUCCESS)
    {
        output += "  CL_PLATFORM_NAME:\t " + std::string(buffer) + "\n";
    }
    else
    {
        std::cerr << "  Error " << status << " in clGetPlatformInfo Call\n\n" << std::endl;
        exit(1);
    }

    status = clGetPlatformInfo(clSelectedPlatformID, CL_PLATFORM_VERSION, sizeof(buffer), buffer, NULL);
    if (status == CL_SUCCESS)
    {
        output += "  CL_PLATFORM_VERSION:\t " + std::string(buffer) + "\n";
    }
    else
    {
        std::cerr << "  Error " << status << " in clGetPlatformInfo Call\n\n" << std::endl;
        exit(1);
    }

    output += "\nDevice Info:\n";

    cl_uint numDevices;
    cl_device_id *devices;

    status = clGetDeviceIDs(clSelectedPlatformID, CL_DEVICE_TYPE_ALL, 0, NULL, &numDevices);
    if (numDevices == 0)
        std::cerr << "  No devices found supporting OpenCL \n" << std::endl;

    if (status != CL_SUCCESS)
    {
        std::cerr << "  Error " << status << " in clGetDeviceIDs Call\n\n" << std::endl;
        exit(1);
    }

    output += "  Found " + ToString(numDevices) + " OpenCL device(s)\n";
    devices = (cl_device_id*)malloc(sizeof(cl_device_id) * numDevices);

    status = clGetDeviceIDs(clSelectedPlatformID, CL_DEVICE_TYPE_ALL, numDevices, devices, &numDevices);
    if (status == CL_SUCCESS)
    {
         for(unsigned int i = 0; i < numDevices; ++i )
         {
             output += " ---------------------------------\n";
             clGetDeviceInfo(devices[i], CL_DEVICE_NAME, sizeof(buffer), &buffer, NULL);
             output += "  Device " + std::string(buffer) + "\n";
             output += " ---------------------------------\n";
             getDeviceInfo(output, devices[i]);
             //oclPrintDevInfo(LOGBOTH, devices[i]);
              //sProfileString += ", Device = ";
             //sProfileString += cBuffer;
         }
     }


    std::cout << output << std::endl;
}

void getProgBinary(cl_program cpProgram, cl_device_id cdDevice, char** binary, size_t* length)
{
    // Grab the number of devices associated witht the program
    cl_uint num_devices;
    clGetProgramInfo(cpProgram, CL_PROGRAM_NUM_DEVICES, sizeof(cl_uint), &num_devices, NULL);

    // Grab the device ids
    cl_device_id* devices = (cl_device_id*) malloc(num_devices * sizeof(cl_device_id));
    clGetProgramInfo(cpProgram, CL_PROGRAM_DEVICES, num_devices * sizeof(cl_device_id), devices, 0);

    // Grab the sizes of the binaries
    size_t* binary_sizes = (size_t*)malloc(num_devices * sizeof(size_t));
    clGetProgramInfo(cpProgram, CL_PROGRAM_BINARY_SIZES, num_devices * sizeof(size_t), binary_sizes, NULL);

    // Now get the binaries
    char** ptx_code = (char**) malloc(num_devices * sizeof(char*));
    for( unsigned int i=0; i<num_devices; ++i) {
        ptx_code[i]= (char*)malloc(binary_sizes[i]);
    }
    clGetProgramInfo(cpProgram, CL_PROGRAM_BINARIES, 0, ptx_code, NULL);

    // Find the index of the device of interest
    unsigned int idx = 0;
    while( idx<num_devices && devices[idx] != cdDevice ) ++idx;

    // If it is associated prepare the result
    if( idx < num_devices )
    {
        *binary = ptx_code[idx];
        *length = binary_sizes[idx];
    }

    // Cleanup
    free( devices );
    free( binary_sizes );
    for( unsigned int i=0; i<num_devices; ++i) {
        if( i != idx ) free(ptx_code[i]);
    }
    free( ptx_code );
}

////////////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////////////

namespace po = boost::program_options;

int main(int argc, char **argv){
    cl_context       clGPUContext;
    cl_program       clProgram;

    size_t           dataBytes;
    size_t           kernelLength;
    cl_int           errcode;

    std::string commandLine;
    for(int i = 0; i < argc; i++)
    {
       commandLine += argv[i];
       commandLine += " ";
    }
    commandLine += "\n";
    //std::cout <<  commandLine << std::endl;

    std::string platform;
    // Declare the supported options.
    po::options_description desc("Usage: oclcc [options] input-file\nOptions");
    desc.add_options()
    ("output-file,o",po::value<std::string >( ), "File for compilation output.\n")
        ("includes,I",po::value<std::vector<std::string > >( ), "Directories to search for headers.\n")
        ("define,D",po::value<std::vector<std::string > >( ), "Defines a macro.\n")
        ("input-file",po::value<std::string >( ), "Input file to compile.\n")
        ("platform",po::value<std::string >(&platform)->default_value("NVIDIA"), "Platform to compile for.\n")
        ("cl-single-precision-constant",     "Treat floating-point constant as single \nprecision constant "
                                             "instead of implicitly converting it to double precision "
                                             "constant. This is valid only when the double precision extension "
                                             "is supported. This is the default if double precision "
                                             "floating-point is not supported.\n")
        ("cl-denorms-are-zero",              "This option controls how single precision and double precision denormalized "
                                             "numbers are handled.\n")
        ("cl-opt-disable",                   "This option disables all optimizations. The default is optimizations are "
                                             "enabled.\n")
        ("cl-strict-aliasing",               "This option allows the compiler to assume the strictest aliasing rules.\n")
        ("cl-mad-enable",                    "Allows a * b + c to be replaced by a mad. This will result in reduced accuracy.\n")
        ("cl-no-signed-zeros",               "Allow optimizations for floating-point \narithmetic that ignore the signedness of zero.\n")
        ("cl-unsafe-math-optimizations",     "Allow optimizations for floating-point \narithmetic that (a) assume that arguments and "
                                             "results are valid, (b) may violate IEEE 754 standard and (c) may violate the OpenCL "
                                             "\nnumerical compliance requirements.\n")
        ("cl-finite-math-only",              "Allow optimizations for floating-point arithmetic that assume that arguments and results "
                                             "are not NaNs or +/- Inf.\n")
        ("cl-fast-relaxed-math",             "Sets the optimization options \n--cl-finite-math-only and \n--cl-unsafe-math-optimizations.\n")
        ("cl-nv-maxrregcount",po::value<std::string >( ), "Nvidia specific: The max number of registers a GPU function can use.\n")
        ("cl-nv-opt-level",po::value<std::string >( ), "Nvidia specific: optimization level (0 - 3).\n")
        ("cl-nv-verbose",                    "Nvidia specific: turns on verbose build\noutput.\n")
        ("warnings-off,w",                   "Turn off warnings.\n")
         ("Werror",                           "Turn warnings into errors.\n")
        ("device-query",                     "Query OpenCL devices and exit.\n")
        ("version",                          "Print out version info and exit.\n")
        ("help", "produce help message");

    po::positional_options_description p;
    p.add("input-file", -1);

    po::variables_map vm;
    try {
    po::store(po::command_line_parser(argc, argv).
          options(desc).positional(p).run(), vm);

    po::notify(vm);
    } catch (po::error & e) {
       std::cerr << std::endl << e.what() << std::endl << std::endl;
       std::cout << desc << "\n";
       exit(1);
    }

    if (vm.count("device-query"))
    {
       deviceQuery();
       return 0;
    }
    if (vm.count("version")) {
        std::cout << version << "\n";
        return 0;
    }
    if (vm.count("help")) {
        std::cout << version << "\n";
        std::cout << info << "\n";
        std::cout << desc << "\n";
        return 0;
    }
    if (!vm.count("input-file"))
    {
       std::cerr << "\nError: must specify input file" << std::endl;
       std::cout << desc << "\n";
       exit(1);
     }
    std::string inFile = vm["input-file"].as<std::string>();

    std::string outFile = "a.out";
    if (vm.count("output-file"))
       outFile = vm["output-file"].as<std::string>();

    std::vector<std::string> includes;
    std::string includesStr;
    if (vm.count("includes"))
       includes = vm["includes"].as<std::vector<std::string> >();
    for(uint i = 0; i < includes.size(); i++)
    {
       includesStr += "-I";
       includesStr += includes[i];
       includesStr += " ";
    }

    std::vector<std::string> defines;
    std::string definesStr;
    if (vm.count("define"))
       defines = vm["define"].as<std::vector<std::string> >();
    for(uint i = 0; i < defines.size(); i++)
    {
       definesStr += "-D";
       definesStr += defines[i];
       definesStr += " ";
    }

    // set build options to be passed into the ocl compile
    std::string buildOptions("");
    buildOptions = definesStr;
    buildOptions += includesStr;

    if (vm.count("Werror") && vm.count("warnings-off"))
    {
       std::cerr << "I think you are confused..." << std::endl;
       std::cerr << "   You can't really want warnings to be turned off AND errors at the same time." << std::endl;
       std::cout << desc << "\n";
       exit(1);
     }

    if (vm.count("Werror"))
       buildOptions += "-Werror ";
    if (vm.count("warnings-off"))
       buildOptions += "-w ";
    if (vm.count("cl-nv-maxrregcount"))
       buildOptions += "-cl-nv-maxrregcount=" + vm["cl-nv-maxrregcount"].as<std::string>() + " ";
    if (vm.count("cl-nv-opt-level"))
       buildOptions += "-cl-nv-opt-level=" + vm["cl-nv-opt-level"].as<std::string>() + " ";
    if (vm.count("cl-nv-verbose"))
       buildOptions += "-cl-nv-verbose ";
    if (vm.count("cl-single-precision-constant"))
       buildOptions += "-cl-single-precision-constant ";
    if (vm.count("cl-denorms-are-zero"))
       buildOptions += "-cl-denorms-are-zero ";
    if (vm.count("cl-opt-disable"))
       buildOptions += "-cl-opt-disable ";
    if (vm.count("cl-strict-aliasing"))
       buildOptions += "-cl-strict-aliasing ";
    if (vm.count("cl-mad-enable"))
       buildOptions += "-cl-mad-enable ";
    if (vm.count("cl-no-signed-zeros"))
       buildOptions += "-cl-no-signed-zeros ";
    if (vm.count("cl-unsafe-math-optimizations"))
       buildOptions += "-cl-unsafe-math-optimizations ";
    if (vm.count("cl-finite-math-only"))
       buildOptions += "-cl-finite-math-only ";
    if (vm.count("cl-fast-relaxed-math"))
       buildOptions += "-cl-fast-relaxed-math ";

    cl_platform_id   cpPlatform;
    cl_device_id     cdDevice;

    errcode = ocltGetPlatformID(&cpPlatform, platform.c_str());
    checkError(errcode, CL_SUCCESS);

    // Get a GPU device
    errcode = clGetDeviceIDs(cpPlatform, CL_DEVICE_TYPE_GPU, 1, &cdDevice, NULL);
    checkError(errcode, CL_SUCCESS);
    
     clGPUContext = clCreateContext(0, 1, &cdDevice, NULL, NULL, &errcode);

    checkError(errcode, CL_SUCCESS);

    // get the list of GPU devices associated with context
    errcode = clGetContextInfo(clGPUContext, CL_CONTEXT_DEVICES, 0, NULL, &dataBytes);
    cl_device_id *cdDevices = (cl_device_id *)malloc(dataBytes);
    errcode |= clGetContextInfo(clGPUContext, CL_CONTEXT_DEVICES, dataBytes, cdDevices, NULL);
    checkError(errcode, CL_SUCCESS);

    // load source and build it
    char *clBinomialOption = ocltLoadKernelSrc(inFile.c_str(), &kernelLength);
    if(clBinomialOption == NULL)
    {
       std::cout << "Error: input file " << inFile << " not found!" << std::endl;
       exit(1);
    }

    clProgram = clCreateProgramWithSource(clGPUContext, 1, (const char **)&clBinomialOption, &kernelLength, &errcode);
    checkError(errcode, CL_SUCCESS);

    errcode = clBuildProgram(clProgram, 0, NULL, buildOptions.c_str(), NULL, NULL);

    // Get the build log... Without doing this it is next to impossible to
    // debug a failed .cl build
    ocltLogBuildInfo(clProgram, cdDevices[0]);
    if(errcode != CL_SUCCESS)
       exit(1);

    // Store the binary in the file system
    char* binary;
    size_t binaryLength;
    getProgBinary(clProgram, cdDevices[0], &binary, &binaryLength);
    checkError(errcode, CL_SUCCESS);
    FILE* fp = fopen(outFile.c_str(), "wb");
    fwrite(binary, 1, binaryLength, fp);

     // append the build options to the binary
    std::string temp = "//!@#~";
    temp += buildOptions;
    fwrite(temp.c_str(), 1, temp.length(), fp);
    fclose(fp);

    errcode |= clReleaseProgram(clProgram);
    errcode |= clReleaseContext(clGPUContext);
    checkError(errcode, CL_SUCCESS);

    free(cdDevices);
}

             
